/**
 * Socket.c
 *
 * Copyright 2001, 2003 by Anthony Howe. All rights reserved.
 */

#undef NDEBUG

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#if defined(__WIN32__)

#	if defined(__BORLANDC__)
#		include <io.h>
#	endif
#	if defined(__VISUALC__)
#		define _WINSOCK2API_
#	endif

#	include <windows.h>
#	include <winsock2.h>

#	define UPDATE_ERRNO	if (errno == 0)	errno = WSAGetLastError()

#	define EINPROGRESS	WSAEINPROGRESS
#	define EISCONN          WSAEISCONN

#	define USE_GETHOSTBYNAME

#include <com/snert/lib/io/Log.h>

#else /* a sensible OS */

#	include <syslog.h>
#	include <sys/time.h>
#	include <sys/socket.h>
#	include <netdb.h>
#	include <arpa/inet.h>
#	include <netinet/in.h>
#	include <netinet/tcp.h>
#	include <fcntl.h>
#	include <unistd.h>

int
WSAGetLastError(void)
{
	return errno;
}

#	define UPDATE_ERRNO

#	ifndef INADDR_NONE
#	define INADDR_NONE	((in_addr_t) -1)
#	endif

#	if defined(__bsdi__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__CYGWIN__)
#		define USE_GETHOSTBYNAME
#	endif

#endif /* a sensible OS */

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Socket.h>

/***********************************************************************
 *** Global Variables, Constants, Macros
 ***********************************************************************/

#define GETHOSTBYNAME_BUFFER_SIZE	2048

struct ipaddress {
	socklen_t length;
	struct sockaddr_in address;
};

struct socket {
	int fd;
	int error;
	int readFlags;
	int writeFlags;
	int onShutdown;
	struct ipaddress send;
	struct ipaddress recv;
};

#define GET_FD(s)		(((struct socket *) s)->fd)

static long debug = 0;

/***********************************************************************
 *** Global Functions
 ***********************************************************************/

void
SocketSetDebugMask(int mask)
{
	debug = mask;
}

/*
 * Called before any SocketOpen() in the parent thread or process.
 */
int
SocketInit(void)
{
	static int SocketSupportLoaded = 0;

	if (!SocketSupportLoaded) {
#if defined(__WIN32__)
/*
 * WinSock 2.2 provides a subset of the BSD Sockets API.
 */
		int rc;
		WORD version;
		WSADATA wsaData;

		version = MAKEWORD(2, 2);
		if ((rc = WSAStartup(version, &wsaData)) != 0) {
			fprintf(stderr, "WSAStartup() failed: %d\n", rc);
			return -1;
		}

		if (HIBYTE( wsaData.wVersion ) < 2 || LOBYTE( wsaData.wVersion ) < 2) {
			fprintf(stderr, "WinSock API must be version 2.2 or better.\n");
			(void) WSACleanup();
			return -1;
		}

		if (atexit((AtExitFunction) WSACleanup)) {
			fprintf(stderr, "atexit(WSACleanup) failed: %s\n", strerror(errno));
			return -1;
		}
#endif
		SocketSupportLoaded = -1;
	}

	return 0;
}

/***********************************************************************
 *** Socket Functions
 ***********************************************************************/

/*
 * Convert ASCII IP address or host name into an in_addr structure.
 * This function is similar to inet_aton, which is not available on
 * all platforms. The difference here is that it will also do
 * hostname lookup. Return 0 on success otherwise -1 on error.
 */
int
SocketAddressToIP(const char *address, struct in_addr *ip)
{
	if (address == (const char *) 0 || ip == (struct in_addr *) 0)
		return -1;

#ifdef MODIFY_AND_IGNORE_CONST
	if (*address == '[') {
		address[strcspn(address, ']')] = '\0';
		address++;
	}
#endif

	/* First try it as aaa.bbb.ccc.ddd. */
#if defined(HAVE_INET_PTON)
	if (inet_pton(AF_INET, address, ip))
		return 0;
#elif defined(HAVE_INET_ATON)
	if (inet_aton(address, ip))
		return 0;
#else
	if ((ip->s_addr = inet_addr(address)) != INADDR_NONE)
		return 0;
#endif

#ifdef USE_GETHOSTBYNAME
{
	struct hostent *host;

	/*** This is NOT thread safe. This should be wrapped by a mutex. ***/
	if ((host = gethostbyname(address)) == (struct hostent *) 0) {
#ifndef NDEBUG
		if (debug & SOCKET_DEBUG_OPEN_CLOSE)
			syslog(LOG_DEBUG, "gethostbyname(%s) failed, h_errno=%d", address, h_errno);
#endif
		return -1;
	}

	*ip = *(struct in_addr *) *host->h_addr_list;

	return 0;
}
#else
{
	char *hbuf;
	int my_errno = 0, rc;
	struct hostent host, *phost;

	if ((hbuf = malloc(GETHOSTBYNAME_BUFFER_SIZE)) == (char *) 0)
		return -1;

#if defined(__linux__)
	rc = gethostbyname_r(address, &host, hbuf, GETHOSTBYNAME_BUFFER_SIZE, &phost, &my_errno);
#elif defined(__sun__)
	/* SunOS just has to be different:
	 *
	 *	struct hostent  *gethostbyname_r(
	 *		const  char  *name,  struct hostent *result, 
	 *		char *buffer, intbuflen, int *h_errnop
	 *	);
	 */
	rc = -( gethostbyname_r(address, &host, hbuf, GETHOSTBYNAME_BUFFER_SIZE, &my_errno) == NULL );
#else
	rc = gethostbyname_r(address, &host, hbuf, GETHOSTBYNAME_BUFFER_SIZE, &my_errno);
#endif
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "gethostbyname_r(%s,...) rc=%d, my_errno=%d", address, rc, my_errno);
#endif
	/* Set the return code to 0 or -1. Note we only
	 * support IPv4 for the moment.
	 */
	rc = -(rc != 0 || my_errno != 0 || host.h_length != sizeof *ip);
	if (rc == 0)
		*ip = *(struct in_addr *) host.h_addr_list[0];

	free(hbuf);

	return rc;
}
#endif
}

Socket
SocketOpen(const char *host, int port, int type)
{
	struct socket *sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpen(%s, %d, %d)", host, port, type);
#endif
	errno = 0;
	sock = (Socket) calloc(1, sizeof *sock);
	if (sock == (Socket) 0)
		goto error0;

	sock->send.length = sizeof sock->send.address;
	sock->send.address.sin_port = htons((unsigned short) port);
	sock->send.address.sin_family = AF_INET;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "call SocketAddressToIP(%s,...)", host);
#endif
	/* Convert server host address from ASCII to binary. */
	if (SocketAddressToIP(host, &sock->send.address.sin_addr))
		goto error1;

	switch (type) {
	case SOCK_STREAM:
		/* By default close only one-half of the connection. */
		SocketSetShutdown(sock, SHUT_WR);
		break;
	case SOCK_DGRAM:
		/* You can't shutdown a UDP connection, duh. */
		SocketSetShutdown(sock, -1);
		break;
	default:
#ifndef NDEBUG
		syslog(LOG_DEBUG, "unsupported socket type %d", type);
#endif
		errno = EINVAL;
		goto error1;
	}

	/* Get us the socket. */
	sock->fd = socket(PF_INET, type, 0);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "socket( ) fd=%d", sock->fd);
#endif
	if (sock->fd < 0) {
		UPDATE_ERRNO;
		goto error1;
	}

#ifdef USE_IP_TOS
	(void) SocketSetTypeOfService(sock, IPTOS_LOWDELAY);
#endif
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpen(%s, %d, %d) %lx", host, port, type, sock);
#endif
	return (Socket) sock;
error1:
	free(sock);
error0:
	return (Socket *) 0;
}

Socket
SocketFromFd(int fd)
{
	struct socket *sock;

	sock = (Socket) calloc(1, sizeof *sock);
	if (sock == (Socket) 0)
		goto error0;

	if (sock->fd < 0)
		goto error1;

	sock->fd = fd;

	return (Socket) sock;
error1:
	free(sock);
error0:
	return (Socket *) 0;
}

Socket
SocketOpenUdp(const char *host, int port)
{
	struct socket *sock;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "enter SocketOpenUdp(%s, %d)", host, port);
#endif
	if ((sock = SocketOpen(host, port, SOCK_DGRAM)) != (Socket) 0) {
		(void) SocketSetLinger((Socket) sock, 1);
	}

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "exit SocketOpenUdp(%s, %d) %lx", host, port, sock);
#endif
	return sock;
}

Socket
SocketOpenTcpClient(const char *host, int port)
{
	struct socket *sock;

	if ((sock = SocketOpen(host, port, SOCK_STREAM)) == (Socket) 0)
		goto error0;

	/* Establish the client to server connection using the "liberal
	 * behaviour" technique for blocking connect() suggested by:
	 *
	 *  http://www.eleves.ens.fr:8080/home/madore/computers/connect-intr.html
	 */
	while (connect(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length) != 0) {
		UPDATE_ERRNO;
		if (errno == EISCONN)
			break;
		if (errno != EINTR)
			goto error1;
	}

	return (Socket) sock;
error1:
	SocketClose(sock);
error0:
	return (Socket) 0;
}

Socket
SocketOpenTcpClientWait(const char *host, int port, long timeout)
{
	int rc;
	struct socket *sock;

	if ((sock = SocketOpen(host, port, SOCK_STREAM)) == (Socket) 0)
		goto error0;

/* man connect
 *
 * EINPROGRESS
 * 	The  socket  is  non-blocking  and the connection cannot be com-
 * 	pleted immediately.  It is possible to select(2) or poll(2)  for
 * 	completion  by  selecting  the  socket for writing. After select
 * 	indicates writability, use getsockopt(2) to  read  the  SO_ERROR
 * 	option  at  level  SOL_SOCKET  to determine whether connect com-
 * 	pleted  successfully  (SO_ERROR  is  zero)   or   unsuccessfully
 * 	(SO_ERROR  is one of the usual error codes listed here, explain-
 * 	ing the reason for the failure).
 */
	rc = SocketSetNonBlocking(sock, 1);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "SocketSetNonBlocking(s, 1) rc=%d errno=%d", rc, errno);
#endif
	if (rc != 0)
		goto error1;

	rc = connect(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length);
	UPDATE_ERRNO;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "connect(%d, ip, len) rc=%d errno=%d so_error=%d", sock->fd, rc, errno, SocketGetError(sock));
#endif
	if (rc != 0 && errno != 0 && errno != EINTR && errno != EINPROGRESS)
		goto error1;

	rc = SocketWaitForOutput(sock, timeout);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "SocketWaitForOutput(s, %ld) rc=%d errno=%d so_error=%d", timeout, rc, errno, SocketGetError(sock));
#endif
	if (rc != 0 || SocketGetError(sock) != 0)
		goto error1;

	rc = SocketSetNonBlocking(sock, 0);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE)
		syslog(LOG_DEBUG, "SocketSetNonBlocking(s, 0) rc=%d errno=%d", rc, errno);
#endif
	if (rc != 0)
		goto error1;

	return (Socket) sock;
error1:
	SocketClose(sock);
error0:
	return (Socket) 0;
}

Socket
SocketOpenTcpServer(const char *host, int port, int nqueue)
{
	struct socket *sock;

	if ((sock = SocketOpen(host, port, SOCK_STREAM)) == (Socket) 0)
		goto error0;

	SocketClearError(sock);
	(void) SocketSetReuseAddr((Socket) sock, 1);

	/* Setup a server socket. Once the socket is open, its the
	 * responsibilty of the application to wait for connections
	 * by calling SocketAccept() or SocketWaitForInput().
	 *
	 * We cannot do that here since the SocketAccept() is often
	 * used within a loop to spawn threads or processes to handle
	 * the client connection.
	 */
	sock->send.address.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(sock->fd, (struct sockaddr *) &sock->send.address, sock->send.length) < 0)
		goto error1;

	if (listen(sock->fd, nqueue) < 0)
		goto error1;

	return (Socket) sock;
error1:
	UPDATE_ERRNO;
	SocketClose(sock);
error0:
	return (Socket *) 0;
}

void
SocketShutdownNow(Socket sock, int shut)
{
	int rc;

	rc = shutdown(GET_FD(sock), shut);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_OPEN_CLOSE) {
		UPDATE_ERRNO;
		syslog(LOG_DEBUG, "immediate socket shutdown(%d, %d) rc=%d errno=%d", GET_FD(sock), shut, rc, errno);
	}
#endif
}

void
SocketClose(Socket sock)
{
	int rc;

	if (sock != (Socket) 0) {
		if (0 <= GET_FD(sock)) {
			if (0 <= ((struct socket *) sock)->onShutdown) {
				rc = shutdown(GET_FD(sock), ((struct socket *) sock)->onShutdown);
#ifndef NDEBUG
				if (debug & SOCKET_DEBUG_OPEN_CLOSE) {
					UPDATE_ERRNO;
					syslog(LOG_DEBUG, "socket shutdown(%d, %d) rc=%d errno=%d", GET_FD(sock), ((struct socket *) sock)->onShutdown, rc, errno);
				}
#endif
			}
#if defined(__WIN32__)
			rc = closesocket(GET_FD(sock));
#else
			rc = close(GET_FD(sock));
#endif
#ifndef NDEBUG
			if (debug & SOCKET_DEBUG_OPEN_CLOSE) {
				UPDATE_ERRNO;
				syslog(LOG_DEBUG, "socket close(%d) rc=%d errno=%d", GET_FD(sock), rc, errno);
			}
#endif
		}

		free(sock);
	}
}

int
ConvertIpToString(const void *ip, char *buffer, long length)
{
#if defined(HAS_INET_NTOP) && defined(AF_INET6)
{
	/* struct in6_addr *ip */
	return -(inet_ntop(AF_INET6, ip, buffer, length) == (char *) 0);
}
#else
{
	/* struct in_addr *ip */
	struct in_addr from;
	(void) memcpy(&from, (unsigned char *) ip, sizeof from);
	(void) strncpy(buffer, inet_ntoa(from), length);
	return 0;
}
#endif
}

static char *
SocketIpToAddress(struct ipaddress *ip)
{
	char addr[50];

	if (ConvertIpToString(&ip->address.sin_addr, addr, sizeof addr))
		return (char *) 0;

	return strdup(addr);
}

char *
SocketGetAddress(Socket sock)
{
	if (sock == (Socket) 0)
		return (char *) 0;

	return SocketIpToAddress(&((struct socket *) sock)->send);
}

char *
SocketGetSourceAddress(Socket sock)
{
	if (sock == (Socket) 0)
		return (char *) 0;

	return SocketIpToAddress(&((struct socket *) sock)->recv);
}

int
SocketGetPort(Socket sock)
{
	return sock == (Socket) 0 ? -1 : ntohs(((struct socket *) sock)->send.address.sin_port);
}

int
SocketGetFd(Socket sock)
{
	return sock == (Socket) 0 ? -1 : GET_FD(sock);
}

Socket
SocketAccept(Socket server)
{
	struct socket *c = calloc(1, sizeof *c);
	struct socket *s = (struct socket *) server;

	SocketClearError(server);

	if (c == (Socket) 0) {
		s->error = ENOMEM;
		goto error0;
	}

	c->send.length = sizeof c->send.address;

	if ((c->fd = accept(s->fd, (struct sockaddr *) &c->send.address, &c->send.length)) < 0)
		goto error1;

	if (SocketSetKeepAlive((Socket) c, 1) < 0)
		goto error1;

	return (Socket) c;
error1:
	(void) SocketGetError(server);
	SocketClose(c);
error0:
	return (Socket) 0;
}

/*
 * Wait for input on a given socket until input is ready or the time
 * to wait given in milliseconds, expires. A zero (0) wait time causes
 * the function to block indefinitely.
 *
 * Return 0 if input ready, else ETIMEDOUT for a timeout, otherwise
 * the value of errno.
 */
int
SocketWaitForInput(Socket sock, long milliseconds)
{
	int rc;
	fd_set readset;
	struct timeval timeout;
	struct socket *s = (struct socket *) sock;

	SocketClearError(sock);

	FD_ZERO(&readset);
	FD_SET((unsigned) s->fd, &readset);

	timeout.tv_sec = milliseconds / 1000L;
	timeout.tv_usec = (milliseconds % 1000L) * 1000L;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ)
		syslog(LOG_DEBUG, "SocketWaitForInput(%d, %ld) waiting...", s->fd, milliseconds);
#endif
	for (;;) {
		rc = select(
			s->fd + 1, &readset, (fd_set *) 0, (fd_set *) 0,
			0 < milliseconds ? &timeout : (struct timeval *) 0
		);

		if (0 < rc) {
			/* Its alive!. */
			(void) SocketGetError(s);
			break;
		}

		if (rc == 0) {
			/* Nothing interesting happened. Explicity set
			 * the error code since the man page does not
			 * state whether errno will be set ETIMEDOUT.
			 */
			(void) SocketGetError(s);
			s->error = ETIMEDOUT;
			break;
		}

		if (errno != EINTR) {
			/* "Oh no! Not again." said the pot of petunias.
			 * Check for a socket error now, otherwise save
			 * errno as our socket error.
			 */
			if (SocketGetError(s) == 0)
				s->error = errno;
			break;
		}
	}

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ)
		syslog(LOG_DEBUG, "SocketWaitForInput(%d, %ld) rc=%d error=%d", s->fd, milliseconds, rc, s->error);
#endif
	return s->error;
}

int
SocketWaitForOutput(Socket sock, long milliseconds)
{
	int rc;
	fd_set writeset;
	struct timeval timeout;
	struct socket *s = (struct socket *) sock;

	SocketClearError(sock);

	FD_ZERO(&writeset);
	FD_SET((unsigned) s->fd, &writeset);

	timeout.tv_sec = milliseconds / 1000L;
	timeout.tv_usec = (milliseconds % 1000L) * 1000L;

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_WRITE)
		syslog(LOG_DEBUG, "SocketWaitForOutput(%d, %ld) waiting...", s->fd, milliseconds);
#endif
	for (;;) {
		rc = select(
			s->fd + 1, (fd_set *) 0, &writeset, (fd_set *) 0,
			0 < milliseconds ? &timeout : (struct timeval *) 0
		);

		if (0 < rc) {
			/* Its alive!. */
			(void) SocketGetError(s);
			break;
		}

		if (rc == 0) {
			/* Nothing interesting happened. Explicity set
			 * the error code since the man page does not
			 * state whether errno will be set ETIMEDOUT.
			 */
			(void) SocketGetError(s);
			s->error = ETIMEDOUT;
			break;
		}

		if (errno != EINTR) {
			/* "Oh no! Not again." said the pot of petunias.
			 * Check for a socket error now, otherwise save
			 * errno as our socket error.
			 */
			if (SocketGetError(s) == 0)
				s->error = errno;
			break;
		}
	}

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_WRITE)
		syslog(LOG_DEBUG, "SocketWaitForOutput(%d, %ld) rc=%d error=%d", s->fd, milliseconds, rc, s->error);
#endif
	return s->error;
}

#ifdef USE_IP_TOS
int
SocketSetTypeOfService(Socket sock, int tos)
{
	return setsockopt(GET_FD(sock), IPPROTO_IP, IP_TOS, (char *) &tos, sizeof tos);
}
#endif

int
SocketSetTcpNoDelay(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_TCP, TCP_NODELAY, (char *) &flag, sizeof flag);
}

int
SocketSetBroadcast(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_BROADCAST, (char *) &flag, sizeof flag);
}

int
SocketSetKeepAlive(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_KEEPALIVE, (char *) &flag, sizeof flag);
}

int
SocketSetReuseAddr(Socket sock, int flag)
{
	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof flag);
}

int
SocketSetLinger(Socket sock, int seconds)
{
	struct linger setlinger = { 1, 0 };

#if defined(__WIN32__)
	setlinger.l_onoff = (u_short) (0 < seconds);
	setlinger.l_linger = (u_short) seconds;
#else
	setlinger.l_onoff = 0 < seconds;
	setlinger.l_linger = seconds;
#endif

	return setsockopt(GET_FD(sock), SOL_SOCKET, SO_LINGER, (char *) &setlinger, sizeof setlinger);
}

void
SocketSetShutdown(Socket sock, int shutdown)
{
	((struct socket *) sock)->onShutdown = shutdown;
}

/*
 * Return zero (0) on success, otherwise -1 on error. Remember
 * that setting this affects both socket input and output.
 */
int
SocketSetNonBlocking(Socket sock, int flag)
{
	int rc;
	long flags;

#if defined(__WIN32__)
	flags = flag;
	rc = ioctlsocket(GET_FD(sock), FIONBIO, (unsigned long *) &flags);
#else
	flags = (long) fcntl(GET_FD(sock), F_GETFL);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_SET)
		syslog(LOG_DEBUG, "fcntl(%d, F_GETFL) flags=%lx errno=%d", GET_FD(sock), flags, errno);
#endif

	if (flag)
		flags |= O_NONBLOCK;
	else
		flags &= ~O_NONBLOCK;

	rc = fcntl(GET_FD(sock), F_SETFL, flags);
#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_SET)
		syslog(LOG_DEBUG, "fcntl(%d, F_SETFL, %lx) rc=%d errno=%d", GET_FD(sock), flags, rc, errno);
#endif
#endif
	return rc == 0 ? 0 : -1;
}

void
SocketSetWriteFlags(Socket sock, int flags)
{
	((struct socket *) sock)->writeFlags = flags;
}

void
SocketSetReadFlags(Socket sock, int flags)
{
	((struct socket *) sock)->readFlags = flags;
}

void
SocketClearError(Socket sock)
{
	errno = 0;
	((struct socket *) sock)->error = 0;
}

int
SocketGetError(Socket sock)
{
	struct socket *s = (struct socket *) sock;
	socklen_t length = sizeof s->error;

	UPDATE_ERRNO;

	if (s->error == 0) {
		int saved_error = errno;

		/* getsockopt(SO_ERROR) resets the socket's copy of the error code,
		 * so we keep it around in our own structure until the next Socket*
		 * operation.
		 */
		(void) getsockopt(s->fd, SOL_SOCKET, SO_ERROR, (void *) &s->error, &length);

		/* SocketClearError() clears both errno and so_error. So if
		 * so_error is zero and errno is not zero after a socket operation,
		 * then assume that errno is our socket error. This assumes that
		 * errno is part if the thread's context.
		 */
		if (s->error == 0 && saved_error != 0)
			s->error = saved_error;
	}

	return s->error;
}

long
SocketWriteIp(Socket sock, void *buffer, long length, IpAddress anIP)
{
	long nbytes;
	struct socket *s = (struct socket *) sock;
	struct ipaddress *ip = (struct ipaddress *) anIP;

	SocketClearError(sock);

	nbytes = sendto(
		s->fd, buffer, (size_t) length, s->writeFlags,
		(const struct sockaddr *) &ip->address, ip->length
	);

	if (nbytes < 0)
		(void) SocketGetError(sock);

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_WRITE)
		syslog(LOG_DEBUG, "SocketWriteIp(%d, %lx, %ld, %lx) nbytes=%ld errno=%d so_error=%d", s->fd, (long) buffer, length, ip, nbytes, errno, s->error);
#endif
	return nbytes;
}

long
SocketWrite(Socket sock, void *buffer, long length)
{
	return SocketWriteIp(sock, buffer, length, &((struct socket *) sock)->send);
}

long
SocketRead(Socket sock, void *buffer, long length)
{
	long nbytes;
	struct socket *s = (struct socket *) sock;

	SocketClearError(sock);

	s->recv.length = sizeof s->recv.address;
	nbytes = recvfrom(
		s->fd, buffer, length, s->readFlags,
		(struct sockaddr *) &s->recv.address, &s->recv.length
	);
	if (nbytes < 0)
		(void) SocketGetError(sock);

#ifndef NDEBUG
	if (debug & SOCKET_DEBUG_READ) {
		int lastbyte = 0 < nbytes ? ((unsigned char *) buffer)[nbytes-1] : -1;
		syslog(
			LOG_DEBUG, "SocketRead(%d, %lx, %ld) nbytes=%ld errno=%d so_error=%d lastbyte=%d %c",
			s->fd, (long) buffer, length, nbytes, errno, s->error,
			lastbyte, isprint(lastbyte) ? lastbyte : ' '
		);
	}
#endif

	return nbytes;
}

long
SocketReadLine(Socket sock, char *line, long size)
{
	long i;
	long nbytes;
	struct socket *s = (struct socket *) sock;

	if (line == (char *) 0 || size <= 0)
		return -1;

	/* Leave room in the buffer for a terminating null byte. */
	line[--size] = '\0';

	for (i = 0; i < size; ++i) {
		/* Keep the buffer null terminated. */
		line[i] = '\0';

		nbytes = SocketRead(sock, &line[i], 1);
		if (nbytes < 0 && s->error != EINTR && s->error != EAGAIN) {
			/* Read error. */
			return -1;
		}

		if (nbytes == 0) {
			/* Buffer underflow (EOF before newline) or EOF. */
			return 0 < i ? i : -2;
		}

		if (line[i] == '\n') {
			/* Newline found. */
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';

			return i;
		}
	}

	/* Buffer overflow. */
	return i;
}

/*
 * Return zero (0) if the entire line was sent,
 * otherwise -1 and check SocketGetError();
 */
int
SocketPrint(Socket sock, char *msg)
{
	size_t length = strlen(msg);
	return SocketWrite(sock, msg, length) == (long) length ? 0 : -1;
}

/*
 * Return zero (0) if the entire line was sent,
 * otherwise -1 and check SocketGetError();
 */
int
SocketPrintLine(Socket sock, char *msg)
{
	if (SocketPrint(sock, msg))
		return -1;

	return SocketWrite(sock, "\r\n", 2) == 2 ? 0 : -1;
}


