/*
 * DebugMemory.c
 *
 * Copyright 2003 by Anthony Howe.  All rights reserved.
 */

/*
 * Default debug memory pool size in kilobytes. Can be controlled
 * at runtime via an environment variable of the same name.
 */
#ifndef DEBUG_MALLOC_POOL_SIZE
#define DEBUG_MALLOC_POOL_SIZE		(2 * 1024)
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define DEBUG_MALLOC_POOL_SIZE_LENGTH	(sizeof ("DEBUG_MALLOC_POOL_SIZE=") - 1)

#include <errno.h>
#include <stdlib.h>
#include <string.h>

extern char **environ;

#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/util/Memory.h>

#ifdef _REENTRANT
# ifdef __unix__
#  include <pwd.h>
#  include <grp.h>
# endif

#include <com/snert/lib/sys/Mutex.h>

static void *memory_mutex;

void
DebugMallocAtExitCleanup()
{
	void *mutex = memory_mutex;
	memory_mutex = (void *) 0;
	MutexDestroy(mutex);
}
#endif

static void *sbrk_list[10];
static void *DebugMallocHeader;
static int already_in_malloc;

void  
DebugFree(void *chunk, char *file, long line)
{	
	void **list, **stop;

#ifdef _REENTRANT
	if (memory_mutex != (void *) 0 && MutexLock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugFree(%lx): MutexLock() failed at %s:%d", chunk, file, line);
		abort();
	}
#endif
	stop = &sbrk_list[sizeof sbrk_list/sizeof sbrk_list[0]];

	for (list = sbrk_list; list < stop; list++) {
		if (*list == chunk) {
			*list = (void *) 0;
			goto done;
		}
	}

	if (MemoryFree(DebugMallocHeader, chunk)) {
		syslog(LOG_ERROR, "DebugFree(%lx): MemoryFree() failed at %s:%d", chunk, file, line);
		abort();
	}
done:
#ifdef _REENTRANT
	if (memory_mutex != (void *) 0 && MutexUnlock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugFree(%lx): MutexUnlock() failed at %s:%d", chunk, file, line);
		abort();
	}
#endif
}

void *
DebugMalloc(long size, char *file, long line)
{
	void *chunk;
	
	if (already_in_malloc) {
		/* Keep track of sbrk() pointers for when we try and free
		 * them. This space is considered lost, but accounted for
		 * since this space should only be allocated during the
		 * initial application start up and freed on exit.
		 */
		void **list, **stop;
		
		stop = &sbrk_list[sizeof sbrk_list/sizeof sbrk_list[0]];
		
		for (list = sbrk_list; list < stop; list++) {
			if (*list == (void *) 0)
				return *list = sbrk(size);
		}
		
		errno = ENOMEM;
		
		return (void *) 0;
	}

#ifdef _REENTRANT
	/* Assumes that the first malloc() doesn't occur in a thread. */
	if (memory_mutex == (void *) 0) {
		already_in_malloc = 1;

		if ((memory_mutex = MutexCreate("DebugMalloc.mutex")) == (void *) 0) {
			syslog(LOG_ERROR, "DebugMalloc(): MutexCreate() failed at %s:%d", file, line);
			abort();
		}

		atexit(DebugMallocAtExitCleanup);

#ifdef __BORLANDC__
		if (MutexPermission(memory_mutex, 0666, 0, 0) < 0) {
#else
		if (MutexPermission(memory_mutex, 0666, getpid(), getgid()) < 0) {
#endif
			syslog(LOG_ERROR, "DebugMalloc(): MutexPermission() failed at %s:%d", file, line);
			abort();
		}
	}

	if (MutexLock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugMalloc(): MutexLock() failed at %s:%d", file, line);
		abort();
	}
#endif	
	
	already_in_malloc = 1;

	if (DebugMallocHeader == (void *) 0) {
		char **env;
		long pool_size;

		/* We have to wait for environ to be initialised before we
		 * can search and allocate a prefered pool size.
		 */
		if (environ == (char **) 0) {
			chunk = DebugMalloc(size, file, line);
			goto done;
		}

		pool_size = DEBUG_MALLOC_POOL_SIZE;
		for (env = environ; *env != (char *) 0; env++) {
			if (strncmp(*env, "DEBUG_MALLOC_POOL_SIZE=", DEBUG_MALLOC_POOL_SIZE_LENGTH) == 0) {
				pool_size = strtol(*env+DEBUG_MALLOC_POOL_SIZE_LENGTH, (char **) 0, 10);
				if (pool_size <= 0)
					pool_size = DEBUG_MALLOC_POOL_SIZE;
				break;
			}
		}	

		DebugMallocHeader = MemoryCreate((void *) 0, pool_size * 1024);
		
		if (DebugMallocHeader == (void *) 0) {
			syslog(LOG_ERROR, "DebugMalloc(): MemoryCreate() failed at %s:%d", file, line);
			abort();
		}
	}
	
	chunk = MemoryAllocate(DebugMallocHeader, size);	
done:
	if (chunk == (void *) 0 && errno == EFAULT) {
		syslog(LOG_ERROR, "DebugMalloc(): MemoryVerifySize() failed at %s:%d", file, line);
		abort();
	}

	already_in_malloc = 0;
	
#ifdef _REENTRANT
	if (MutexUnlock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugMalloc(): MutexUnlock() failed at %s:%d", file, line);
		abort();
	}
#endif	
	return chunk;
}

void *
DebugCalloc(long n, long size, char *file, long line)
{
	void *chunk;
	
	if ((chunk = DebugMalloc(n * size, file, line)) != (void *) 0)
		memset(chunk, 0, n * size);
		
	return chunk;
}

void *
DebugRealloc(void *chunk, long size, char *file, long line)
{
	void *replace;
	
	if (DebugMallocHeader == (void *) 0)
		return DebugMalloc(size, file, line);

#ifdef _REENTRANT
	if (MutexLock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugRealloc(): MutexLock() failed at %s:%d", file, line);
		abort();
	}
#endif
	replace = MemoryReallocate(DebugMallocHeader, chunk, size);
	
	if (replace == (void *) 0 && errno == EFAULT) {
		syslog(LOG_ERROR, "DebugRealloc(): MemoryVerifySize() failed at %s:%d", file, line);
		abort();
	}

#ifdef _REENTRANT
	if (MutexUnlock(memory_mutex)) {		
		syslog(LOG_ERROR, "DebugRealloc(): MutexUnlock() failed at %s:%d", file, line);
		abort();
	}
#endif
	return replace;
}


#ifndef DEBUG_MALLOC_TRACE_CALLER
void 
free(void *p)
{
	DebugFree(p, __FILE__, __LINE__);
}

void *
malloc(size_t n)
{
	return DebugMalloc((long) n, __FILE__, __LINE__);
}

void *
calloc(size_t m, size_t n)
{
	return DebugCalloc(m, (long) n, __FILE__, __LINE__);
}

void *
realloc(void *p, size_t n)
{
	return DebugRealloc(p, (long) n, __FILE__, __LINE__);
}
#endif
