/*
 * EmailSpan.c
 *
 * Assorted span functions for validation.
 *
 * Copyright 2002, 2003 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include <com/snert/lib/util/EmailSpan.h>

/*
 * RFC 2821 section 4.1.3 Address Literals
 *
 *	IPv4-address-literal = Snum 3("." Snum)
 *
 *	Snum = 1*3DIGIT         ; representing a decimal integer
 *				; value in the range 0 through 255
 */
long
EmailSpanIPv4(char *ip)
{
	int dots;
	long octet;
	char *start, *stop;

	dots = 0;
	for (start = ip; *ip != '\0'; ip = stop) {
		octet = strtol(ip, &stop, 10);

		/* Must be an octet between 0..255. */
		if (ip == stop)
			break;

		if (octet < 0 || 255 < octet)
			return ip - start;

		/* Count the dot separators. */
		if (*stop == '.') {
			++stop;
			++dots;
		}
	}

	if (dots != 3)
		return 0;

	return ip - start;
}

long
EmailSpanAddressLiteral(char *s)
{
	char *t = s;

	if (*t != '[')
		return 0;

	if (isdigit(*++t)) {
		t += EmailSpanIPv4(t);
	} else {
		/*** TODO IPv6 address forms. ***/
	}

	if (*t == ']')
		t++;

	return t - s;
}

/*
 * RFC 2821 section 4.1.2 Command Argument Syntax
 *
 * Validate the characters and syntax.
 *
 *     Mailbox = Local-part "@" Domain
 *     Domain = (sub-domain 1*("." sub-domain)) / address-literal
 *     sub-domain = Let-dig [Ldh-str]
 *     Let-dig = ALPHA / DIGIT
 *     Ldh-str = *( ALPHA / DIGIT / "-" ) Let-dig
 */
long
EmailSpanDomainName(char *s, int minimumDots)
{
	char *t;
	int dots;

	if (*s == '[')
		return EmailSpanAddressLiteral(s);

	dots = 0;

	for (t = s; *t != '\0'; t++) {
		/* First charcter of a domain segment must be alpha-numeric. */
		if (!isalnum(*t))
			break;

		/* Rest of domain segment can be alpha-numeric or hypen. */
		for (t++; *t != '\0'; t++) {
			if (!isalnum(*t) && *t != '-')
				break;
		}

		/* Something other than the domain segment delimiter? */
		if (*t != '.')
			break;

		/* Last character of domain segment must be alpha-numeric. */
		if (t[-1] == '-') {
			t--;
			break;
		}

		dots++;
	}

	/* 0	simple machine name		local@machine
	 * 1	a domain name			local@domain.tld
	 * 2	fully qualified domain name	local@machine.domain.tld
	 */
	if (dots < minimumDots)
		return 0;

	return t - s;
}

/*
 * RFC 2822 section 3.2.4. Atom
 *
 * Validate only the characters.
 */
long
EmailSpanLocalPart(char *s)
{
	int i;
	char *t;
	char x[2] = { 0, 0 };

	for (t = s; *t != '\0'; t++) {
		if (isalnum(*t))
			continue;

		*x = *t;
		/* atext does not include '.', but we do here to simplify
		 * scanning dot-atom.
		 */
		i = strspn(x, "!#$%&'*+-/=?^_`{|}~.") ;
		if (i != 1)
			break;
	}

	return t - s;
}

