/*
 * Text.c
 *
 * Copyright 2001, 2003 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <com/snert/lib/util/Buf.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/io/posix.h>

char *
TextDup(const char *orig)
{
	char *copy;

	copy = malloc(strlen(orig) + 1);
	if (copy == (char *) 0)
		return (char *) 0;

	return strcpy(copy, orig);
}

void
boundOffsetLength(long slength, long *offset, long *length)
{
	/* Correct offset if out of bounds. */
	if (*offset < 0)
		*offset = 0;
	else if (slength < *offset)
		*offset = slength;

	/* Correct length if out of bounds. */
	if (*length < 0)
		*length = slength;
	if (slength < *offset + *length)
		*length = slength - *offset;
}

void
TextLower(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) tolower(*s);
	}
}

void
TextUpper(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) toupper(*s);
	}
}

void
TextReverse(char *s, long offset, long length)
{
	long slength;
	char ch, *x, *y;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	/* Reverse segment of string. */
	for (x = y = s + offset, y += length; x < --y; ++x) {
		ch = *y;
		*y = *x;
		*x = ch;
	}
}

void
TextInvert(char *s, long offset, long length)
{
	char *stop;

	if (s == (char *) 0)
		return;
	
	boundOffsetLength(strlen(s), &offset, &length);

	for (stop = s + offset + length; s < stop; s++) {
		*s = (char) (isupper(*s) ? tolower(*s) : toupper(*s));
	}
}

/**
 * <p>
 * Copy at most tlength bytes or until a null byte is seen
 * from the source C string to the target C string. The target
 * string will be null terminated if tlength remains positive.
 * </p>
 *
 * @param t
 *	The target string.
 *
 * @param s
 *	The source string.
 *
 * @return
 *	The number of bytes actually copied, excluding the
 *	null terminator. This will be the length of s or
 *	tlength if the end of the target string buffer was
 *	reached.
 */
long
TextCopyN(char *t, char *s, long tlength)
{
	char *start;

	for (start = s; *s != '\0' && 0 < tlength--; )
		*t++ = *s++;

	if (0 < tlength)
		*t = '\0';

	return s - start;
}

/**
 * <p>
 * Compute the hash value of a string.
 * </p>
 *
 * @param hash
 *	A previously computed hash value.
 *
 * @param s
 *	A string to hash with the previous hash.
 *
 * @return
 *	A new hash value.
 */
long
TextHash(long hash, const char *s)
{
	while (*s != '\0')
		hash = 37 * hash + *s++;

	return hash;
}

/**
 * <p>
 * Is the string empty or blank?  This is equivalent to the
 * Perl regular expression:
 * </p>
 * <pre>
 *	m!^\s*!.
 * </pre>
 *
 * @param s
 *	A string to scan.
 *
 * @return
 *	True if the string is empty (zero length) or contains
 *	only whitespace.
 */
int
TextIsEmpty(const char *s)
{
	return s[strcspn(s, " \t\r\n\f")] == '\0';
}

/**
 * <p>
 * Does the string consist only of digits in the given radix?
 * </p>
 *
 * @param s
 *	A string to scan.
 *
 * @param radix
 *	The expected radix.
 *
 * @return
 *	True if the string conatins only characters for the given
 *	integer in the given radix.
 */
int
TextIsInteger(const char *s, int radix)
{
	char *digit;
	static char digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";

	if (*s == '\0' || radix < 2 || 36 < radix)
		return 0;

	for ( ; *s != '\0'; ++s) {
		digit = strchr(digits, *s);
		if (digit == (char *) 0 || radix <= digit - digits)
			return 0;
	}

	return 1;
}

/**
 * <p>
 * Given the character following a backslash, return the
 * the character's ASCII escape value.
 * </p>
 * <pre>
 *   bell            \a	0x07
 *   backspace       \b	0x08
 *   escape          \e	0x1b
 *   formfeed        \f	0x0c
 *   linefeed        \n	0x0a
 *   return          \r	0x0d
 *   space           \s	0x20
 *   tab             \t	0x09
 *   vertical-tab    \v	0x0b
 * </pre>
 *
 * @param ch
 *	A character that followed a backslash.
 *
 * @return
 *	The ASCII value of the escape character or the character itself.
 */
int
TextBackslash(char ch)
{
	switch (ch) {
	case 'a': return '\007';
	case 'b': return '\010';
	case 'e': return '\033';
	case 'f': return '\014';
	case 'n': return '\012';
	case 'r': return '\015';
	case 's': return '\040';
	case 't': return '\011';
	case 'v': return '\013';
	}

	return ch;
}

/**
 * <p>
 * Parse the string for the next token. A token consists of characters
 * not found in the set of delimiters. It may contain backslash-escape
 * sequences, which shall be converted into literals or special ASCII
 * characters. It may contain single or double quoted strings, in which
 * case the quotes shall be removed, though any backslash escape
 * sequences within the quotes are left as is.
 * </p>
 *
 * @param
 *	A quoted string.
 *
 * @return
 *	An allocated token.
 *
 * @see #TextBackslash(char)
 */
char *
TextToken(const char *string, const char **stop, const char *delims)
{
	char *token, *t;
	const char *s;
	int quote = 0, escape = 0;

	/* Advance to start of token. */
	string += strspn(string, delims);

	/* No more tokens? */
	if (*string == '\0') {
		if (stop != (const char **) 0)
			*stop = string;
		return (char *) 0;
	}

	/* Find end of token. */
	for (s = string; *s != '\0'; ++s) {
		if (escape) {
			escape = 0;
			continue;
		}

		switch (*s) {
		case '"': case '\'':
			quote = *s == quote ? 0 : *s;
			continue;
		case '\\':
			escape = 1;
			if (quote == 0) continue;
			continue;
		}

		if (quote == 0 && strchr(delims, *s) != (char *) 0)
			break;
	}

	token = malloc((s - string) + 1);
	if (token == (char *) 0)
		return (char *) 0;

	/* Copy token, removing quotes and backslashes. */
	for (t = token; string < s; ++string) {
		if (escape) {
			if (quote == 0)
				*t++ = (char) TextBackslash(*string);
			else
				*t++ = *string;
			escape = 0;
			continue;
		}

		switch (*string) {
		case '"': case '\'':
			quote = *string == quote ? 0 : *string;
			continue;
		case '\\':
			escape = 1;
			if (quote == 0) continue;
			break;
		}

		if (quote == 0 && strchr(delims, *string) != (char *) 0)
			break;

		*t++ = *string;
	}
	*t = '\0';

	if (stop != (const char **) 0)
		*stop = s;

	return token;
}

/**
 * <p>
 * The given string contains a list of substrings separated by the
 * specified delimiter characters. The substrings may contain quoted
 * strings and/or contain backslash-escaped characters. The common
 * backslash escape sequences are supported and return their ASCII
 * values.
 * </p>
 *
 * @param string
 *	A list represented as a string.
 *
 * @param delims
 *	A set of delimiter characters.
 *
 * @return
 *	A vector of C strings.
 */
Vector
TextSplit(const char *string, const char *delims)
{
	char *token;
	Vector list;
	
	if ((list = VectorCreate(5)) == (Vector) 0)
		return (Vector) 0;

	VectorOnRemove(list, VectorRemoveTrivial, (void *) 0);

	while ((token = TextToken(string, &string, delims)) != (char *) 0)
		VectorAdd(list, token);

	return list;
}

/**
 * <p>
 * Join a vector of strings into a string with the given delimiter.
 * </p>
 *
 * @param delim
 *	A delimiter of one or more characters.
 *
 * @param strings
 *	A list of strings to be joined.
 *
 * @return
 *	A string comprised of all the objects in string representation
 *	each separated by the string delimiter.
 */
char *
TextJoin(const char *delim, Vector strings)
{
	long i;
	char *s;
	Buf collector = BufCreate(100);

	for (i = 0; i < VectorLength(strings); ++i) {
		BufAddString(collector, VectorGet(strings, i));
		BufAddString(collector, (char *) delim);
	}

	BufSetLength(collector, BufLength(collector) - strlen(delim));
	s = BufToString(collector);
	BufDestroy(collector);

	return s;
}

/**
 * <p>
 * Expand all the tab characters into spaces. Tabstops are assumed
 * to be every 8th column.
 * </p>
 *
 * @param s
 *	A string buffer containing tab characters to be expanded.
 *
 * @param col
 *	Initial column position
 *
 * @return
 * 	A string buffer where all tab characters have been converted
 *	to spaces.
 */
Buf
TextExpand(Buf s, long col)
{
	long i;
	unsigned char byte;
	Buf t = BufCreate(100);

	for (i = 0; i < BufLength(s); ++i, ++col) {
		byte = BufGetByte(s, i);
		if (byte == '\t') {
			do {
				BufAddByte(t, ' ');
			} while ((++col & 7) != 0);
		} else {
			BufAddByte(t, byte);
		}
	}

	return t;
}

/**
 * <p>
 * Convert a byte array into a hex encoded string.
 * </p>
 *
 * @param b
 *	A byte array.
 *
 * @return
 *	A string representing the byte array as hexdecimal values.
 */
char *
TextHexEncode(Buf b)
{
	long i;
	Buf s = BufCreate(BufLength(b) * 2 + 1);
	static char xdigits[] = "0123456789ABCDEF";

	for (i = 0; i < BufLength(b); ++i) {
		unsigned char a = BufGetByte(b, i);
		BufAddByte(s, xdigits[(a >> 4) & 0xf]);
		BufAddByte(s, xdigits[a & 0xf]);
	}

	BufAddByte(s, '\0');

	return (char *) BufAsBytes(s);
}

int
TextInsensitiveCompareN(const void *xp, void const *yp, long len)
{
	int diff;
	char *x = (char *) xp;
	char *y = (char *) yp;

	if (x == (char *) 0 && y != (char *) 0)
		return 1;
	if (x != (char *) 0 && y == (char *) 0)
		return -1;
	if (x == (char *) 0 && y == (char *) 0)
		return 0;

	for ( ; *x != '\0' && *y != '\0'; ++x, ++y) {
		if (0 <= len && len-- == 0)
			return 0;

		if (*x != *y) {
			diff = tolower(*x) - tolower(*y);
			if (diff != 0)
				return diff;
		}
	}

	if (len == 0)
		return 0;

	return *x - *y;
}

int
TextInsensitiveCompare(const void *xp, void const *yp)
{
	return TextInsensitiveCompareN(xp, yp, -1);
}

/**
 * <p>
 * Compare the leading part of the string in a case insensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param prefix
 *	The string prefix to match.
 *
 * @return
 * 	Return zero (0) on no match, otherwise the length of the
 *	matching prefix.
 */
int
TextInsensitiveStartsWith(const char *text, const char *prefix)
{
	const char *start = text;

	for ( ; *prefix != '\0'; ++text, ++prefix) {
		if (tolower(*text) != tolower(*prefix))
			return 0;
	}

	return (int) (text - start);
}

int
TextInsensitiveEndsWith(const char *text, const char *suffix)
{
	const char *etext = text + strlen(text);
	const char *esuffix = suffix + strlen(suffix);

	while (suffix < esuffix) {
		if (etext <= text)
			return 0;

		--etext;
		--esuffix;

		if (tolower(*etext) != tolower(*esuffix))
			return 0;
	}

	return 1;
}

int
TextSensitiveCompareN(const void *xp, void const *yp, long len)
{
	char *x = (char *) xp;
	char *y = (char *) yp;

	if (x == (char *) 0 && y != (char *) 0)
		return 1;
	if (x != (char *) 0 && y == (char *) 0)
		return -1;
	if (x == (char *) 0 && y == (char *) 0)
		return 0;

	if (len < 0)
		return strcmp(x, y);

	return strncmp(x, y, len);
}

int
TextSensitiveCompare(const void *xp, void const *yp)
{
	return TextSensitiveCompareN(xp, yp, -1);
}

/**
 * <p>
 * Compare the leading part of the string in a case sensitive manner.
 * </p>
 *
 * @param text
 *	The string to check.
 *
 * @param prefix
 *	The string prefix to match.
 *
 * @return
 * 	Return zero (0) on no match, otherwise the length of the
 *	matching prefix.
 */
int
TextSensitiveStartsWith(const char *text, const char *prefix)
{
	const char *start = text;

	for ( ; *prefix != '\0'; ++text, ++prefix) {
		if (*text != *prefix)
			return 0;
	}

	return (int) (text - start);
}

int
TextSensitiveEndsWith(const char *text, const char *suffix)
{
	const char *etext = text + strlen(text);
	const char *esuffix = suffix + strlen(suffix);

	while (suffix < esuffix) {
		if (etext <= text)
			return 0;
		if (*--etext != *--esuffix)
			return 0;
	}

	return 1;
}

/**
 * <p>
 * Read a line of input until a newline (CRLF or LF) is read or the buffer
 * is filled. The newline is removed from the input and the buffer is always
 * null terminated.
 * </p>
 *
 * @param fp
 *	The FILE * from which to read input.
 *
 * @param line
 *	The input buffer.
 *
 * @param size
 *	The size of the input buffer.
 *
 * @return
 * 	Return the length of the input buffer. otherwise -1 on error.
 */
int
TextInputLine(FILE *fp, char *line, int size)
{
	int i;

	for (i = 0, --size; i < size; ++i) {
		line[i] = (char) fgetc(fp);

		if (feof(fp) || ferror(fp))
			return -1;

		if (line[i] == '\n') {
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';
			break;
		}
	}

	line[i] = '\0';

	return i;
}

/**
 * <p>
 * Read a line of input until a newline (CRLF or LF) is read or the buffer
 * is filled. The newline is removed from the input and the buffer is always
 * null terminated.
 * </p>
 *
 * @param fd
 *	The file handle from which to read input.
 *
 * @param line
 *	The input buffer.
 *
 * @param size
 *	The size of the input buffer.
 *
 * @return
 * 	Return the length of the input buffer. otherwise -1 on error.
 */
int
TextReadLine(int fd, char *line, int size)
{
	int i;

	for (i = 0, --size; i < size; ++i) {
		if (read(fd, &line[i], 1) <= 0)
			return -1;

		if (line[i] == '\n') {
			line[i] = '\0';
			if (0 < i && line[i-1] == '\r')
				line[--i] = '\0';
			break;
		}
	}

	line[i] = '\0';

	return i;
}

int
TextCountOccurences(char *s1, char *s2)
{
	char *t1, *t2;
	int count = 0;

	for ( ; *s1 != '\0'; ++s1) {
		for (t1 = s1, t2 = s2; *t2 != '\0'; ++t1, ++t2) {
			if (*t1 != *t2)
				break;
		}
		count += (*t2 == '\0');
	}

	return count;
}

